/*
 * Copyright (c) 2006-2007 Extreme! Lab, Indiana University. All rights reserved.
 *
 * This software is open source. See the bottom of this file for the license.
 *
 * $Id: LEADWorkflowInvoker.java,v 1.22 2007/06/08 02:45:46 sshirasu Exp $
 */
package edu.indiana.extreme.xbaya.lead;

import java.net.URI;
import java.util.List;
import java.util.regex.Pattern;

import javax.xml.namespace.QName;

import org.xmlpull.v1.builder.XmlElement;

import xsul.XmlConstants;
import xsul.invoker.puretls.PuretlsInvoker;
import xsul.lead.LeadContextHeader;
import xsul.wsif.WSIFMessage;
import xsul.wsif.WSIFOperation;
import xsul.wsif.WSIFPort;
import xsul.wsif.WSIFService;
import xsul.wsif.WSIFServiceFactory;
import xsul.wsif.impl.WSIFMessageElement;
import xsul.wsif.spi.WSIFProviderManager;
import xsul.wsif_xsul_soap_gsi.Provider;
import xsul.xhandler_soap_sticky_header.StickySoapHeaderHandler;
import xsul.xwsif_runtime.WSIFClient;
import xsul.xwsif_runtime.WSIFRuntime;
import xsul.xwsif_runtime_async.WSIFAsyncResponsesCorrelator;
import xsul.xwsif_runtime_async_http.XsulSoapHttpWsaResponsesCorrelator;
import xsul.xwsif_runtime_async_msgbox.XsulMsgBoxWsaResponsesCorrelator;
import xsul5.MLogger;
import xsul5.wsdl.WsdlDefinitions;
import edu.indiana.extreme.xbaya.XBayaException;
import edu.indiana.extreme.xbaya.component.ComponentException;
import edu.indiana.extreme.xbaya.component.ws.WSComponent;
import edu.indiana.extreme.xbaya.component.ws.WSComponentFactory;
import edu.indiana.extreme.xbaya.component.ws.WSComponentPort;
import edu.indiana.extreme.xbaya.util.WSDLUtil;
import edu.indiana.extreme.xbaya.util.XMLUtil;

/**
 * @author Satoshi Shirasuna
 */
public class LEADWorkflowInvoker {

    private static final MLogger logger = MLogger.getLogger();

    private WsdlDefinitions definitions;

    private WSIFClient client;

    private WSIFOperation operation;

    private boolean success;

    private WSIFMessage inputMessage;

    private WSIFMessage outputMessage;

    private WSIFMessage faultMessage;

    private WSComponent component;

    private LeadContextHeader leadContext;

    private URI messageBoxURL;

    private PuretlsInvoker secureInvoker;

    static {
        WSIFProviderManager.getInstance().addProvider(
                new xsul.wsif_xsul_soap_http.Provider());
    }

    /**
     * Constructs a LEADWorkflowInvoker.
     * 
     * @param definitions
     * @param leadContext
     * @throws ComponentException
     */
    public LEADWorkflowInvoker(WsdlDefinitions definitions,
            LeadContextHeader leadContext) throws ComponentException {
        this(definitions, leadContext, null, (PuretlsInvoker) null);
    }

    /**
     * Constructs a LEADWorkflowInvoker.
     * 
     * @param definitions
     * @param leadContext
     * @param messageBoxURL
     * @throws ComponentException
     */
    public LEADWorkflowInvoker(WsdlDefinitions definitions,
            LeadContextHeader leadContext, URI messageBoxURL)
            throws ComponentException {
        this(definitions, leadContext, messageBoxURL, (PuretlsInvoker) null);
    }

    /**
     * Constructs a LEADWorkflowInvoker.
     * 
     * @param definitions
     * @param leadContext
     * @param messageBoxURL
     * @param secureInvoker
     * @throws ComponentException
     */
    public LEADWorkflowInvoker(WsdlDefinitions definitions,
            LeadContextHeader leadContext, URI messageBoxURL,
            PuretlsInvoker secureInvoker) throws ComponentException {
        this.definitions = definitions;
        this.leadContext = leadContext;
        this.messageBoxURL = messageBoxURL;
        this.secureInvoker = secureInvoker;
        init();
    }

    /**
     * @return The lead context
     */
    public LeadContextHeader getLeadContext() {
        return this.leadContext;
    }

    /**
     * @return The inputs
     */
    public List<WSComponentPort> getInputs() {
        return this.component.getInputPorts();
    }

    /**
     * @param inputs
     */
    public void setInputs(List<WSComponentPort> inputs) {
        for (WSComponentPort input : inputs) {
            String name = input.getName();
            Object value = input.getValue();

            // Give a special treatment for "arrays"
            if (this.inputMessage instanceof WSIFMessageElement) {
                QName type = ((WSIFMessageElement) this.inputMessage)
                        .getPartType(name);
                if (LEADTypes.isArrayType(type)) {
                    // split string into items using " " as separator
                    Pattern pattern = Pattern.compile("[,\\s]+");
                    // String[] result = p.split("one,two, three four , five");
                    String[] result = pattern.split((String) value);
                    XmlElement arrayEl = XmlConstants.BUILDER.newFragment(name);
                    for (int i = 0; i < result.length; i++) {
                        logger.finest("split=" + result[i]);
                        arrayEl.addElement("value").addChild(result[i]);
                    }
                    this.inputMessage.setObjectPart(name, arrayEl);
                    value = null; // no need to set string value below
                }

            }
            if (value != null) {
                this.inputMessage.setObjectPart(name, value);
            }
        }
    }

    /**
     * @return true if it succeeds; false otherwise
     * @throws XBayaException
     */
    public boolean invoke() throws XBayaException {
        try {
            logger.info("leadContext: "
                    + XMLUtil.xmlElementToString(this.leadContext));
            logger
                    .info("inputMessage: "
                            + XMLUtil
                                    .xmlElementToString((XmlElement) this.inputMessage));
            // Soap11Util.getInstance().wrapBodyContent(
            // (XmlElement) this.inputMessage);
            this.success = this.operation.executeRequestResponseOperation(
                    this.inputMessage, this.outputMessage, this.faultMessage);
            if (this.success) {
                logger
                        .info("outputMessage: "
                                + XMLUtil
                                        .xmlElementToString((XmlElement) this.outputMessage));
            } else {
                logger
                        .info("faultMessage: "
                                + XMLUtil
                                        .xmlElementToString((XmlElement) this.faultMessage));
            }
            return this.success;
        } catch (RuntimeException e) {
            String message = "Error in invoking a service.";
            throw new XBayaException(message, e);
        }
    }

    /**
     * @return The outputs
     */
    public List<WSComponentPort> getOutputs() {
        List<WSComponentPort> outputs = this.component.getOutputPorts();
        for (WSComponentPort output : outputs) {
            String name = output.getName();
            String value = this.outputMessage.getObjectPart(name).toString();
            output.setValue(value);
        }
        return outputs;
    }

    /**
     * @return The fault
     */
    public WSIFMessage getFault() {
        return getFaultMessage();
    }

    /**
     * @return true if success; false otherwise.
     */
    public boolean isSuccess() {
        return this.success;
    }

    /**
     * @return the input message.
     */
    public WSIFMessage getInputMessage() {
        return this.inputMessage;
    }

    /**
     * @return the output message.
     */
    public WSIFMessage getOutputMessage() {
        return this.outputMessage;
    }

    /**
     * @return the fault message.
     */
    public WSIFMessage getFaultMessage() {
        return this.faultMessage;
    }

    /**
     * Parses the WSDL and prepares everything.
     * 
     * @throws ComponentException
     */
    private void init() throws ComponentException {
        logger.finest("wsdl: " + this.definitions.xmlStringPretty());
        this.component = WSComponentFactory.createComponent(this.definitions);

        WSIFServiceFactory factory = WSIFServiceFactory.newInstance();
        WSIFService service = factory.getService(WSDLUtil
                .wsdlDefinitions5ToWsdlDefintions3(this.definitions));

        if (this.secureInvoker != null) {
            Provider secureProvider = new xsul.wsif_xsul_soap_gsi.Provider(
                    this.secureInvoker);
            service.addLocalProvider(secureProvider);
        }

        this.client = WSIFRuntime.getDefault().newClientFor(service, null);
        // null selects the first port in the first service.

        // Async
        int clientPort = 0;
        WSIFAsyncResponsesCorrelator correlator;
        if (this.messageBoxURL == null) {
            logger.info("starting response correlator using local port "
                    + clientPort);
            XsulSoapHttpWsaResponsesCorrelator wasCorrelator = new XsulSoapHttpWsaResponsesCorrelator(
                    clientPort);
            String serverLoc = (wasCorrelator).getServerLocation();
            logger.finest("client is waiting at " + serverLoc);
            correlator = wasCorrelator;
        } else {
            logger.info("starting reponse correlator using message box "
                    + this.messageBoxURL);
            correlator = new XsulMsgBoxWsaResponsesCorrelator(
                    this.messageBoxURL.toString());
        }

        this.client.useAsyncMessaging(correlator);

        StickySoapHeaderHandler handler = new StickySoapHeaderHandler(
                "use-lead-header", this.leadContext);
        this.client.addHandler(handler);

        // This handler has to be end to get the entire soap message.
        NotificationHandler notificationHandler = new NotificationHandler(
                this.leadContext);
        this.client.addHandler(notificationHandler);

        WSIFPort port = this.client.getPort();
        String operationName = this.component.getOperationName();
        logger.finest("operationName: " + operationName);
        this.operation = port.createOperation(operationName);
        this.inputMessage = this.operation.createInputMessage();
        this.outputMessage = this.operation.createOutputMessage();
        this.faultMessage = this.operation.createFaultMessage();
    }
}

/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 * 
 * Copyright (c) 2006-2007 The Trustees of Indiana University. All rights
 * reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 * 
 * 2) All redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 * 
 * 3) Any documentation included with all redistributions must include the
 * following acknowledgement:
 * 
 * "This product includes software developed by the Indiana University Extreme!
 * Lab. For further information please visit http://www.extreme.indiana.edu/"
 * 
 * Alternatively, this acknowledgment may appear in the software itself, and
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name "Indiana University" or "Indiana University Extreme! Lab" shall
 * not be used to endorse or promote products derived from this software without
 * prior written permission from Indiana University. For written permission,
 * please contact http://www.extreme.indiana.edu/.
 * 
 * 5) Products derived from this software may not use "Indiana University" name
 * nor may "Indiana University" appear in their name, without prior written
 * permission of the Indiana University.
 * 
 * Indiana University provides no reassurances that the source code provided
 * does not infringe the patent or any other intellectual property rights of any
 * other entity. Indiana University disclaims any liability to any recipient for
 * claims brought by any other entity based on infringement of intellectual
 * property rights or otherwise.
 * 
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS.
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS",
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION
 * GENERATED USING SOFTWARE.
 */

